# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Simple parser for MIP tables.

My interpretation of the format from reading the CMIP5 tables.

"""

import re
from glob import glob
import csv

import logging
log = logging.getLogger(__name__)

entry_ids = ['axis_entry', 'variable_entry']

line_rexp = re.compile(r'(\w+):\s*(.*)')
dquote = '"(?:""|[^"])*"'
squote = "'(?:''|[^'])*'"
val_rexp = re.compile('%s|%s|!|[^!"\']+' % (dquote, squote))
expt_id_ok_rexp = re.compile('(?P<sep1>[\'"])(?P<desc>.*?)(?P=sep1)\s*(?P<sep2>[\'"])(?P<id>.*?)(?P=sep2)')

class error(Exception):
    pass

def parse_line(line):
    # Strip out comment
    line, comment = split_comment(line.strip())
    if line:
        mo = line_rexp.match(line)
        if not mo:
            raise error('Unrecognised line: %s' % line)
        entry, value = mo.groups()
        value = value.strip()
    else:
        entry = value = None

    return (entry, value, comment)


def iter_table(fh):
    """
    Generates events (entry, value, comment) by reading a MIP table from a
    file object.

    """
    for line in fh:
        line = line.strip()
        if line:
            yield parse_line(line)

def iter_entries(fh):
    """
    Generate events (entry_name, value_dict) by reading a MIP table from a
    file object.

    """
    entry_type = 'global'
    entry_name = None
    d = {}
    for name, value, comment in iter_table(fh):
        if name is None: 
            continue
        if name in entry_ids:
            yield (entry_type, entry_name, d)
            entry_type = name
            entry_name = value
            d = {}
        else:
            d.setdefault(name, []).append(value)

    yield (entry_type, entry_name, d)


def split_comment(line):
    """
    Detect comment.

    Quoted '!' characters are detected.

    """
    parts = val_rexp.findall(line)
    try:
        i = parts.index('!')
    except ValueError:
        value = ''.join(parts)
        comment = None
    else:
        value = ''.join(parts[:i])
        comment = ''.join(parts[i:])

    return (value, comment)

class MIPTable(object):
    """
    Hold information from a MIP table.  

    This information is used to enforce DRS vocabularies.

    :property name: The name of the MIP table as used in DRS filenames.
    :property variables: A list of variables in this table.
    :property experiments: A list of valid experiment ids for this table.

    """
    def __init__(self, filename):
        """
        :param filename: Name of file containing the MIP table.

        """
        fh = open(filename)
        
        self._vardict = {}

        self._read_entries(fh)
        self._init_experiments()

    def _read_entries(self, fh):
        for entry_type, entry_name, d in iter_entries(fh):
            if entry_type == 'global':
                self.name = re.match('Table (.*)', d['table_id'][0]).group(1)
                self._globals = d
            elif entry_type == 'variable_entry':
                self._vardict[entry_name] = d
            else:
                # Ignore other entry types
                pass

    def _init_experiments(self):
        self._exptdict = {}
        for value in self._globals.get('expt_id_ok', []):
            mo = expt_id_ok_rexp.match(value)
            if not mo:
                raise error("Error parsing expt_id_ok value %s" % value)
            self._exptdict[mo.group('id')] = mo.group('desc')
            

    @property
    def variables(self):
        return self._vardict.keys()

    @property
    def experiments(self):
        return self._exptdict.keys()
    
    @property
    def frequency(self):
        try:
            return self._globals['frequency'][0]
        except KeyError:
            raise AttributeError()

    def get_variable_attr(self, variable, attr):
        """
        Retrieve an attribute of variable.

        If the attributes isn't in the variable entry the global
        value is returned

        """
        if variable not in self._vardict:
            raise ValueError('Variable %s not found' % variable)

        try:
            return self._vardict[variable][attr]
        except KeyError:
            return self.get_global_attr(attr)
            
    def get_global_attr(self, attr):
        try:
            return self._globals[attr]
        except KeyError:
            raise AttributeError('Attribute %s is not a global entry' % attr)

class MIPTableStore(object):
    """
    Holds a collection of mip tables.

    :property tables: A mapping of table names to IMIPTable instances

    """

    def __init__(self, table_glob):
        """
        :param table_glob: A wildcard pattern for all MIP tables to load.

        """
        self.tables = {}

        for filename in glob(table_glob):
            self.add_table(filename)

    def add_table(self, filename):
        """
        Read filename as a MIP table and add it to the store.
        
        :return: The added MIPTable instance.

        """
        t = MIPTable(filename)
        log.info('Adding table %s from %s to table store' % (t.name, filename))
        self.tables[t.name] = t

        return t

    def get_variable_attr(self, table, variable, attr):
        """
        Return the value of a variable's attribute in a given table.

        """
        v = self.get_variable_attr_mv(table, variable, attr)
        if len(v) != 1:
            raise ValueError('%s is a multi-valued MIP attribute' % v)

        return v[0]

    def get_variable_attr_mv(self, table, variable, attr):
        """
        Return the value of a variable's attribute in a given table.

        """
        if table not in self.tables:
            raise ValueError('Table %s not found' % table)

        return self.tables[table].get_variable_attr(variable, attr)
        
    def get_global_attr(self, table, attr):
        """
        Return global table attribute.

        """
        v = self.get_global_attr_mv(table, attr)
        if len(v) != 1:
            raise ValueError('%s is a multi-valued MIP attribute' % v)

        return v[0]

    def get_global_attr_mv(self, table, attr):
        """
        Return the value of a variable's attribute in a given table.

        """
        if table not in self.tables:
            raise ValueError('Table %s not found' % table)

        return self.tables[table].get_global_attr(attr)


    #!FIXME
    #def get_variable_tables(self, variable, **attribute_constraints):
    #    ret = []
    #    for table in self.tables.values():
    #        if variable in table.variables:
    #            for k, v in attribute_constraints.items():
    #                try:
    #                    if table.get_variable_attr(variable, k) != v:
    #                        break
    #                except AttributeError:
    #                    pass
    #            else:
    #                ret.append(table.name)
    #    
    #    return ret


def read_model_table(table_csv):
    """
    Read Karl's CMIP5_models.xls file in CSV export format and 
    return a map of institute to model name.

    This function is invoked internally to load CMIP5_models.xls from inside
    drslib.

    """
    fh = open(table_csv)
    table_reader = csv.reader(fh)

    # Check first 2 lines look like the right file
    header1 = table_reader.next()
    header2 = table_reader.next()

    assert "CMIP5 Modeling Groups" in header1[0]
    assert 'Abbreviated name of center or group' in header2[1]
    assert "modified model_id" in header2[4]

    model_map = {}
    for row in table_reader:
        institute = row[1]
        model = row[4]

        # If institute contains a "/" take the first item
        if '/' in institute:
            institute = institute.split('/')[0]

        if model in model_map:
            raise "Duplicate model key %s" % model
        model_map[model] = institute

    return model_map



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Manage DRS directory structure versioning.

This module provides an API for manipulating a DRS directory structure
to facilitate keeping multiple versions of datasets on disk
simultaniously.  The class :class:`DRSTree` provides a top-level
interface to the DRS directory structure and a container for :class:`PublisherTree` objects.

:class:`PublisherTree` objects expose the versions present in a
publication-level dataset and what files are unversioned.  Calling
:meth:`PublisherTree.do_version` will manipulate the directory structure
to move unversioned files into a new version.

Detailed diagnostics can be logged by setting handlers for the logger
``drslib.drs_tree``.


"""

import os, sys
from glob import glob
import stat
import datetime
import re

from drslib.cmip5 import make_translator
from drslib.translate import TranslationError
from drslib.drs import DRS, path_to_drs, drs_to_path
from drslib import config, mapfile
from drslib.p_cmip5 import ProductScope
from drslib.publisher_tree import PublisherTree

import logging
log = logging.getLogger(__name__)

# We also want to log to p_cmip5 so that product detection can be filtered sensibly
p_cmip5_log = logging.getLogger('drslib.p_cmip5')


class DRSTree(object):
    """
    Manage a Data Reference Syntax directory structure.

    A DRSTree represents the root of a DRS hierarchy.  Also associated
    with DRSTree objects is a incoming directory pattern that is
    searched for files matching the DRS structure.  Any file within
    the incoming tree will be considered for new versions of PublisherTrees.

    :ivar incoming: :class:`DRSList` of (filepath, DRS) of all files to be added to the
                    DRSTree on next upgrade.
    :ivar incomplete: :class:`DRSList` of (filepath, DRS) of all files rejected because
                      of incomplete DRS attributes.

    """

    def __init__(self, drs_root, table_store=None):
        """
        :param drs_root: The path to the DRS *activity* directory.
        :param table_store: Override the default table store.  This can be used
            to select the TAMIP tables.

        """
        self.drs_root = drs_root
        self.pub_trees = {}
        self._vtrans = make_translator(drs_root, table_store=table_store)
        self.incoming = DRSList()
        self.incomplete = DRSList()
        self._p_cmip5 = None

        self._move_cmd = config.move_cmd

        if not os.path.isdir(drs_root):
            raise Exception('DRS root "%s" is not a directory' % self.drs_root)

    def discover(self, incoming_dir=None, **components):
        """
        Scan the directory structure for PublisherTrees.

        To prevent an exaustive scan of the directory structure some
        components of the DRS must be specified as keyword arguments
        or configured via *metaconfig*.  These components are
        *product*, *institute* and *model* 

        The components *experiment*, *frequency* and *realm* are
        optional.  All components can be set to wildcard values.  This
        allows an exaustive scan to be forced if desired.

        :incoming_dir: A filesystem wildcard which should resolve to 
            directories to recursively scan for files.  If None no incoming
            files are detected

        """
        
        drs_t = DRS(**components)

        # NOTE: None components are converted to wildcards
        pt_glob = drs_to_path(self.drs_root, drs_t)
        pub_trees = glob(pt_glob)
        for pt_path in pub_trees:
            # Detect whether pt_path is inside incoming.  If so ignore.
            if incoming_dir and (os.path.commonprefix((pt_path+'/', incoming_dir+'/')) == incoming_dir+'/'):
                log.warning("PublisherTree path %s is inside incoming, ignoring" % pt_path)
                continue

            drs = path_to_drs(self.drs_root, pt_path, activity=drs_t.activity)
            drs_id = drs.to_dataset_id()
            if drs_id in self.pub_trees:
                raise Exception("Duplicate PublisherTree %s" % drs_id)

            log.info('Discovered PublisherTree at %s' % pt_path)
            self.pub_trees[drs_id] = PublisherTree(drs, self)

        # Scan for incoming DRS files
        if incoming_dir:
            self.discover_incoming(incoming_dir, **components)


        
    def discover_incoming(self, incoming_dir, **components):
        """
        Scan the filesystem for incoming DRS files.

        This method can be repeatedly called to discover incoming
        files independently of :meth:`DRSTree.discover` repeatedly .

        :incoming_dir: A directory to recursively scan for files.

        """

        def _iter_incoming():
            for dirpath, dirnames, filenames in os.walk(incoming_dir):
                for filename in filenames:
                    yield (filename, dirpath)

        self.discover_incoming_fromfiles(_iter_incoming(), **components)


    def discover_incoming_fromfiles(self, files_iter, **components):
        """
        Process a stream of files into the incoming list from an
        iterable.

        This method is useful as a low-level hook for integrating
        with processing pipelines.

        :files_iter: An iterable of (filename, path) for
            each file to process into incoming.

        """
        for filename, dirpath in files_iter:
            log.debug('Processing %s' % filename)
            try:
                drs = self._vtrans.filename_to_drs(filename)
            except TranslationError:
                # File doesn't match
                log.warn('File %s is not a DRS file' % filename)
                continue

            log.debug('File %s => %s' % (repr(filename), drs))
            for k, v in components.items():
                if v is None:
                    continue
                # If component is present in drs act as a filter
                drs_v = drs.get(k, None)
                if drs_v is not None:
                    if drs_v != v:
                        log.warn('FILTERED OUT: %s.  %s != %s' %
                                  (drs, repr(drs_v), repr(v)))
                        break
                else:
                    # Otherwise set as default
                    log.debug('Set %s=%s' % (k, repr(v)))
                    setattr(drs, k, v)
            else:
                # Only if break not called

                # Detect product if enabled
                if self._p_cmip5:
                    self._detect_product(dirpath, drs)

                if drs.is_publish_level():
                    log.debug('Discovered %s as %s' % (filename, drs))
                    self.incoming.append((os.path.join(dirpath, filename), drs))
                else:
                    log.debug('Rejected %s as incomplete %s' % (filename, drs))
                    self.incomplete.append((os.path.join(dirpath, filename), drs))
                # Instantiate a PublisherTree for each unique publication-level dataset

        for path, drs in self.incoming:
            drs_id = drs.to_dataset_id()
            if drs_id in self.pub_trees:
                self.pub_trees[drs_id].deduce_state()
            else:
                self.pub_trees[drs_id] = PublisherTree(drs, self)


            
    def remove_incoming(self, path):
        # Remove path from incoming
        #!TODO: This isn't efficient.  Refactoring of incoming or _todo required.
        for npath, drs in self.incoming:
            if path == npath:
                self.incoming.remove((npath, drs))
                break
        else:
            # not found
            raise Exception("File %s not found in incoming" % path)

    def set_p_cmip5(self, p_cmip5):
        """
        Set the :class:`p_cmip5.product.cmip5_product` instance used to deduce
        the DRS product component.

        """
        self._p_cmip5 = p_cmip5

    def _detect_product(self, path, drs):
        """
        Use the p_cmip5 module to deduce the product of this DRS object.
        p_cmip5 must be configured by calling :meth:`DRSTree.set_p_cmip5`.

        """
        p_cmip5_log.info('Deducing product for %s' % drs)

        pci = self._p_cmip5
        if drs.subset is None or drs.subset[0] == None:
            startyear = None
        else:
            startyear = drs.subset[0][0]

        try:
            status = pci.find_product(drs.variable, drs.table, drs.experiment, drs.model,
                                      path, startyear=startyear)
            # Make sure status is consistent with no exceptions being raised
            assert status
        except ProductScope as e:
            p_cmip5_log.warn('FAILED product detection for %s, %s' % (drs, e))
        else:
            p_cmip5_log.debug('%s, %s, %s, %s:: %s %s' % (drs.variable, drs.table, drs.experiment, 
                                                                 path, 
                                                                 pci.product, pci.reason ))
            
            drs.product = pci.product
            p_cmip5_log.info('Product deduced as %s, %s' % (drs.product, pci.reason))

    def set_move_cmd(self, cmd):
        self._move_cmd = cmd

    def incomplete_dataset_ids(self):
        """
        Return a set of dataset ids for each publication-level dataset that detect_incoming()
        has found as incomplete.
        
        """
        if self.incomplete is None:
            return set()
        else:
            return set(drs.to_dataset_id() for fp, drs in self.incomplete)

class DRSList(list):
    """
    A list of tuples (filepath, DRS) objects offering a simple query interface.

    """

    def select(self, **kw):
        """Select all DRS objects with given component values.
        
        For each key in ``kw`` if the value is a list select all
        DRS objects with values in that list, otherwise select
        all objects with that value.

        """
        items = self
        for k, v in kw.items():
            if type(v) == list:
                items = [x for x in items if x[1].get(k, None) in v]
            else:
                items = [x for x in items if x[1].get(k, None) == v]

        return DRSList(items)

        




# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.




# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

from setuptools import setup, find_packages
import sys, os

version = '0.3.0'
drs_version = '0.28'

setup(name='drslib',
      version=version,
      description="A library for processing the CMIP5 Data Reference Syntax",
      long_description="""\

This library supports the generation of paths and filenames
corresponding to version %(drs_version)s of the CMIP Data Reference
Syntax [DRS]_.

.. [DRS] http://cmip-pcmdi.llnl.gov/cmip5/docs/cmip5_data_reference_syntax.pdf

""" % globals(),
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='Stephen Pascoe',
      author_email='Stephen.Pascoe@stfc.ac.uk',
      url='http://esgf.org/esgf-drslib-site/',
      download_url='http://ndg.nerc.ac.uk/dist',
      license='BSD',
      packages=find_packages(exclude=['ez_setup', 'examples', 'test']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'metaconfig',
        # For cmip5_product_identifier
        'xlrd',
      ],
      #tests_require=['NoseXUnit'],
      entry_points= {
        'console_scripts': ['translate_cmip3 = drslib.translate_cmip3:main',
                            'drs_tool = drslib.drs_command:main',
                            'drs_checkthredds = drslib.thredds:main'],
        },
      test_suite='nose.collector',
      )



"""
Check metadata in a THREDDS catalog is consistent with the DRS.

Things to check:

1. All DRS components set as properties and validate with drslib
2. drs_id is consistent with properties
3. version is a date
4. dataset urlPath is consistent with the DRS directory structure.
5. Checksums are present and the right format (NOT 'MD5:...')
6. tracking_id is present
7. Check product assignement is right.

Currently implemented: 1-3

"""

import sys
import os
import re

from lxml import etree as ET
from drslib.drs import DRS
from drslib.cmip5 import make_translator
import urlparse
from optparse import OptionParser

import logging
log = logging.getLogger(__name__)

THREDDS_NS = 'http://www.unidata.ucar.edu/namespaces/thredds/InvCatalog/v1.0'
XLINK_NS = 'http://www.w3.org/1999/xlink'

usage = """%prog [options] thredds ...

thredds:
  A thredds url or file path of a dataset published by esgpublish.\
"""



trans = make_translator('')
drs_prop_map = {'dataset_version': 'version',
                'project': 'activity',
                'experiment': 'experiment',
                'product': 'product',
                'model': 'model',
                'time_frequency': 'frequency',
                'realm': 'realm',
                'cmor_table': 'table',
                'ensemble': 'ensemble',
                'institute': 'institute'
                }


class InvalidThreddsException(Exception):
    """
    An exception raised to indicate failure of a ThreddsCheck
    """
    pass

class CheckNotPossible(Exception):
    """
    Raised to indicate a check doesn't have enough information to continue.

    """
    pass

class ThreddsCheck(object):
    """
    Base class of all checks, defining the interface.

    """

    def __init__(self, environ=None):
        """
        environ is a dictionary shared accross all checks allowing
        checks to share information.
        """
        self.environ = environ


    def check(self, etree):
        """
        Check the THREDDS catalogue represented by an ElementTree object.

        """
        pass



def run_checks(etree, checks, environ=None):
    """
    Run a sequence of checks on a THREDDS catalogue as an ElementTree.
    InvalidThreddsExceptions are converted to log messages.
 
    """
    if environ is None:
        environ = {}

    for CheckClass in checks:
        check = CheckClass(environ)
        try:
            check.check(etree)
        except InvalidThreddsException, e:
            log.error(e)
        except CheckNotPossible:
            log.warn('Check %s aborted' % CheckClass.__name__)
        else:
            log.info('Check %s succeeded' % CheckClass.__name__)

    return environ

class DRSIdCheck(ThreddsCheck):
    """
    Check drs_id is present and consistent with dataset_id.

    """

    def check(self, etree):
        dataset = get_dataset(etree)

        drs_id = get_property(dataset, 'drs_id')
        dataset_id = get_property(dataset, 'dataset_id')

        # Check 2 ids are consistent
        if drs_id != dataset_id:
            raise InvalidThreddsException("dataset_id != drs_id for dataset %s" %
                                          dataset.get('ID'))

        self.environ['dataset_id'] = dataset_id
        self.environ['drs_id'] = drs_id


class DRSPropCheck(ThreddsCheck):
    """
    Check all drs components are defined as properties.

    Creates a drs attribute in the environment if successful.

    """
    def check(self, etree):
        dataset = get_dataset(etree)

        props = {}
        for prop_name in drs_prop_map:
            prop = get_property(dataset, prop_name)
            if prop_name is 'dataset_version':
                prop = int(prop)
            elif prop_name is 'ensemble':
                #!TODO: refactor this to share code with drslib.translate
                mo = re.match(r'(?:r(\d+))?(?:i(\d+))?(?:p(\d+))?', prop)
                if not mo:
                    raise InvalidThreddsException('Unrecognised ensemble syntax %s' % prop)

                (r, i, p) = mo.groups()
                prop = tuple(x and int(x) for x in (r, i, p))

            props[drs_prop_map[prop_name]] = prop

        drs = DRS(**props)

        # If present in environ check against drs_id
        if 'drs_id' in self.environ:
            if drs.to_dataset_id() != self.environ['drs_id']:
                raise InvalidThreddsException("drs properties inconsistent with drs_id for dataset %s" %
                                              dataset.get('ID'))

        self.environ['drs'] = drs


class ValidDRSCheck(ThreddsCheck):
    """
    Check the drs object in the environment is valid.

    """

    def check(self, etree):
        if 'drs' not in self.environ:
            raise CheckNotPossible

        drs = self.environ['drs']
        
        try:
            path = trans.drs_to_path(drs)
        except:
            raise InvalidThreddsException("drs %s fails to validate" % drs)


class ValidDateCheck(ThreddsCheck):
    """
    Check date versioning.

    """
    def check(self, etree):
        if not 'drs' in self.environ:
            raise CheckNotPossible

        drs = self.environ['drs']
        if not drs.version > 20100101:
            raise InvalidThreddsException("The version of dataset doesn't look like a date: %s" %
                                          drs)

#
# Utility functions
#
def get_dataset(etree):
    # There should be only 1 top-level dataset element
    datasets =  etree.findall('{%s}dataset' % THREDDS_NS)
    if len(datasets) != 1:
        raise InvalidThreddsException("More than one top-level dataset")

    return datasets[0]


def get_property(dataset, name):
    prop = dataset.find('{%s}property[@name="%s"]' % 
                      (THREDDS_NS, name))
    if prop is None:
        raise InvalidThreddsException("Property %s not found in dataset %s" % 
                                      (name, dataset.get('ID')))
                                      
    return prop.get('value')


def read_master_catalog(catalog_url):
    """
    Read master catalogue and generate dataset catalogue ElementTree objects.

    """
    cat_etree = ET.parse(catalog_url)
    scheme, netloc, path, query, fragment = urlparse.urlsplit(catalog_url)
    base_url = urlparse.urlunsplit((scheme, netloc, os.path.dirname(path)+'/', None, None))

    for catalog_ref in cat_etree.findall('{%s}catalogRef' % THREDDS_NS):
        ds_url = catalog_ref.get('{%s}href' % XLINK_NS)
        abs_ds_url = urlparse.urljoin(base_url, ds_url)
        
        yield abs_ds_url


def main(argv=sys.argv):
    logging.basicConfig(level=logging.ERROR)

    checks = [DRSIdCheck, DRSPropCheck, ValidDRSCheck, ValidDateCheck]

    op = OptionParser(usage)
    op.add_option('-c', '--catalog', action='store',
                  help="Scan root THREDDS catalog CATALOG for catalogRef "
                        "elements and check each referenced catalog")

    opts, args = op.parse_args(argv[1:])

    xmls = args
    if opts.catalog:
        log.info('Discovering catalogs from master catalog %s' % opts.catalog)
        xmls += list(read_master_catalog(opts.catalog))

    if not xmls:
        op.print_help()

    for xml in xmls:
        log.info('Checking %s' % xml)
        etree = ET.parse(xml)
        run_checks(etree, checks)

if __name__ == '__main__':
    main()
        




# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

import os, sys
from glob import glob
import stat
import datetime
import re

from drslib.cmip5 import make_translator
from drslib.translate import TranslationError
from drslib.drs import DRS, path_to_drs, drs_to_path
from drslib import config, mapfile
from drslib.p_cmip5 import ProductScope

import logging
log = logging.getLogger(__name__)

#---------------------------------------------------------------------------
# DRY definitions

VERSIONING_FILES_DIR = 'files'
VERSIONING_LATEST_DIR = 'latest'
IGNORE_FILES_REGEXP = r'^\..*'

#---------------------------------------------------------------------------

class PublisherTree(object):
    """
    A directory tree at the publication level.

    :cvar STATE_INITIAL: Flag representing the initial unversioned state

    :cvar STATE_VERSIONED: Flag representing the fully versioned state
        with no unversioned files

    :cvar STATE_VERSIONED_TRANS: Flag representing the partially versioned state
        with unversioned files outstanding.

    :param latest: Integer version number of latest version or 0

    """
    #!TODO: At some point we want to check incoming files to see if they are
    #       duplicates of already versioned files.

    STATE_INITIAL = 'INITIAL'
    STATE_VERSIONED = 'VERSIONED'
    STATE_VERSIONED_TRANS = 'VERSIONED_TRANS'

    CMD_MOVE = 0
    CMD_LINK = 1
    CMD_MKDIR = 2

    DIFF_NONE = 0
    DIFF_TRACKING_ID = 1
    DIFF_SIZE = 2
    DIFF_V1_ONLY = 4
    DIFF_V2_ONLY = 8
    DIFF_PATH = 16

    def __init__(self, drs, drs_tree):

        self.drs_tree = drs_tree
        self.drs = drs
        self.state = None
        self._todo = []
        self.versions = {}
        self.latest = 0
        self._vtrans = make_translator(drs_tree.drs_root)
        self._cmortrans = make_translator(drs_tree.drs_root, with_version=False)

        #!TODO: calling internal method.  Make this method public.
        ensemble = self.drs._encode_ensemble()
        self.pub_dir = os.path.join(self.drs_tree.drs_root,
                                      self.drs.product,
                                      self.drs.institute,
                                      self.drs.model,
                                      self.drs.experiment,
                                      self.drs.frequency,
                                      self.drs.realm,
                                      self.drs.table,
                                      ensemble)
        self.deduce_state()

    def deduce_state(self):
        """
        Scan the directory structure to work out what state the
        tree is in.

        """

        self._deduce_versions()
        self._deduce_todo()

        if not self.versions:
            self.state = self.STATE_INITIAL
        elif self._todo:
            self.state = self.STATE_VERSIONED_TRANS
        else:
            self.state = self.STATE_VERSIONED


    def do_version(self, next_version=None):
        """
        Move incoming files into the next version

        """

        self._setup_versioning()

        if next_version is None:
            next_version = self._next_version()

        log.info('Transfering %s to version %d' % (self.pub_dir, next_version))
        self._do_commands(self.todo_commands(next_version))
        self.deduce_state()
        self._do_latest()

    def list_todo(self, next_version=None):
        """
        Return an iterable of command descriptions in the todo list.

        Each item in the iterable is a unix command-line string of the
        form ``"mv ... ..."`` or ``"ln -s ... ..."``

        """

        if next_version is None:
            next_version = self._next_version()

        for cmd, src, dest in self.todo_commands(next_version):
            if cmd == self.CMD_MOVE:
                yield "%s %s %s" % (self.drs_tree._move_cmd, src, dest)
            elif cmd == self.CMD_LINK:
                yield "ln -s %s %s" % (src, dest)
            elif cmd == self.CMD_MKDIR:
                yield "mkdir -p %s" % (dest, )
            else:
                raise Exception("Unrecognised command type")

    def count_todo(self):
        """
        Return the number of files to be processed on next upgrade.

        """
        return len(self._todo)

    def todo_size(self):
        """
        Return the total size of files in the todo list.

        """
        count = 0
        for filename, drs in self._todo:
            count += os.stat(filename)[stat.ST_SIZE]

        return count

    def list_files(self, version=None):
        """
        Returns a list of netcdf files found under ``version`` directory.
        Where ``version`` is an integer > 0 or None to signify latest.
        """

        if version is None:
            version = self.latest
        if version == 0:
            return []

        if version not in self.versions:
            raise Exception("When searching for NetCDF files you need to provide a `version` argument as an integer or thestring 'latest'.")

        return (filepath for filepath, drs in self.versions[version])

    def count(self, version=None):
        return len(list(self.list_files(version=version)))

    def size(self, version=None):
        count = 0
        for filename in self.list_files(version=version):
            count += os.stat(filename)[stat.ST_SIZE]

        return count

    def diff_version(self, v1, v2=None, by_tracking_id=False):
        """
        Deduce the difference between two versions or between a version
        and the todo list.
        """
        
        files1 = {}
        for filepath, drs in self.versions[v1]:
            files1[os.path.basename(filepath)] = filepath

        if v2 is None:
            fl = self._todo
        else:
            fl = self.versions[v2]

        files2 = {}
        for filepath, drs in fl:
            files2[os.path.basename(filepath)] = filepath

        for file in set(files1.keys() + files2.keys()):
            if file in  files1 and file in files2:
                yield (self._diff_file(files1[file], files2[file], 
                                       by_tracking_id),
                       files1[file], files2[file])
            elif file in files1:
                yield (self.DIFF_V1_ONLY, files1[file], None)
            else:
                yield (self.DIFF_V2_ONLY, None, files2[file])


    def version_to_mapfile(self, version, fh=None):
        if fh is None:
            fh = sys.stdout

        if version not in self.versions:
            raise Exception("Version %d not present in PublisherTree %s" % (version, self.pub_dir))

        mapfile.write_mapfile(self.versions[version], fh)

    def version_drs(self, version=None):
        """
        Return a DRS object representing the PublisherTree at the given version number.
        If the version doesn't exist an exception is raised.

        """

        if version is None:
            version = self.latest
        
        # If unversioned just return the bare DRS
        if version == 0:
            return self.drs

        if version not in self.versions:
            raise Exception("Version %d not present in PublsherTree %s" % (version, self.pub_dir))

        return DRS(self.drs, version=version)

    def todo_commands(self, next_version=None):
        """
        Yield a sequence of tuples (CMD, SRS, DEST) indicating the
        files that need to be moved and linked to transfer to next version.

        CMD is one of self.CMD_MOVE, self.CMD_LINK and self.CMD_MKDIR.  
        CMD_MOVE 
          implies executing the command DRSTree._move_command to move, copy or 
          otherwise relocate the file.
        CMD_LINK 
          implies symbolically linking.
        CMD_MKDIR 
          implies creating a leaf directory DEST and all necessary intermediate 
          directories.  CMD_MKDIR commands are only yielded if the directory 
          does not exist.  SRC is None.

        """
        
        if not self._todo:
            return

        if next_version is None:
            next_version = self._next_version()
        
        done = set()
        for filepath, drs in self._todo:
            filename = os.path.basename(filepath)
            fdir = '%s_%d' % (drs.variable, next_version)
            newpath = os.path.abspath(os.path.join(self.pub_dir, VERSIONING_FILES_DIR,
                                                   fdir, filename))


            # Detect directories needing creation
            ddir_src = os.path.dirname(newpath)
            if not os.path.exists(ddir_src):
                yield self.CMD_MKDIR, None, ddir_src

            yield self.CMD_MOVE, filepath, newpath

            linkpath = os.path.abspath(os.path.join(self.pub_dir, 'v%d' % next_version,
                                                    drs.variable,
                                                    filename))
            # Detect directories needing creation
            ddir_dst = os.path.dirname(linkpath)
            if not os.path.exists(ddir_dst):
                yield self.CMD_MKDIR, None, ddir_dst
            # Make relative to source path ddir_dst
            newpath = os.path.relpath(newpath, ddir_dst)

            yield self.CMD_LINK, newpath, linkpath

            done.add(filename)

        #!TODO: Handle deleted files!

        # Now scan through previous version to find files to update
        if self.latest != 0:
            for filepath, drs in self.versions[self.latest]:
                filename = os.path.basename(filepath)
                if filename not in done:
                    # Find link target from previous version
                    prevlink = os.path.abspath(os.path.join(self.pub_dir, 'v%d' % self.latest,
                                                            drs.variable, filename))
                    pfilepath = os.path.abspath(os.path.join(os.path.dirname(prevlink),
                                                              os.readlink(prevlink)))

                    linkpath = os.path.abspath(os.path.join(self.pub_dir, 'v%d' % next_version,
                                                            drs.variable, filename))
                    
                    # Detect directories needing creation
                    ddir = os.path.dirname(linkpath)
                    if not os.path.exists(ddir):
                        yield self.CMD_MKDIR, None, ddir
                    pfilepath = os.path.relpath(pfilepath, ddir)
                    yield self.CMD_LINK, pfilepath, linkpath


    #-------------------------------------------------------------------
    
    def _do_latest(self):
        version = max(self.versions.keys())
        latest_dir = 'v%d' % version
        log.info('Setting latest to %s' % latest_dir)
        latest_lnk = os.path.join(self.pub_dir, VERSIONING_LATEST_DIR)

        if os.path.exists(latest_lnk):
            os.remove(latest_lnk)
        os.symlink(latest_dir, latest_lnk)

    
    def _do_commands(self, commands):
        for cmd, src, dest in commands:
            if cmd == self.CMD_MOVE:
                self._do_mv(src, dest)
            elif cmd == self.CMD_LINK:
                self._do_link(src, dest)
            elif cmd == self.CMD_MKDIR:
                self._do_mkdir(dest)
            else:
                raise Exception('Internal error: Unrecognised command type %s' % cmd)

    def _do_mv(self, src, dest):
        cmd = '%s %s %s' % (self.drs_tree._move_cmd, src, dest)
        if os.path.exists(dest):
            log.warn('Overwriting existing file: %s' % cmd)
        else:
            log.info(cmd)
        #!TODO: Trap output!
        status = os.system(cmd)
        if status != 0:
            log.warn('System call failed: %d' % status)

        # Remove src from incoming
        self.drs_tree.remove_incoming(src)

    def _do_link(self, src, dest):
        if os.path.exists(dest):
            log.warning('Moving symlink %s' % dest)
            os.remove(dest)

        log.info('Linking %s %s' % (src, dest))

        os.symlink(src, dest)

    def _do_mkdir(self, ddir):
        if not os.path.exists(ddir):

            log.info('Creating %s' % ddir)
            os.makedirs(ddir)
        else:
            log.warning('Directory already exists %s' % ddir)



    def _setup_versioning(self):
        """
        Do initial configuration of directory tree to support versioning.

        """
        if not os.path.exists(self.pub_dir):
            log.info("New PublisherTree being created at %s" % self.pub_dir)
            os.makedirs(self.pub_dir)

        path = os.path.join(self.pub_dir, VERSIONING_FILES_DIR)
        if not os.path.exists(path):
            log.info('Initialising %s for versioning.' % self.pub_dir)
            os.mkdir(path)

    def _next_version(self):
        if config.version_by_date:
            today = datetime.date.today()
            return int(today.strftime('%Y%m%d'))
        else:
            return self.latest+1

    def _deduce_versions(self):
        if config.version_by_date:
            return self._deduce_date_versions()
        else:
            return self._deduce_old_versions()
            
    def _deduce_date_versions(self):
        self.latest = 0
        # Bail out if pub_dir doesn't exist yet.
        if not os.path.exists(self.pub_dir):
            return
        # Detect version paths and sort by date
        for basename in os.listdir(self.pub_dir):
            if not re.match(r'v\d+$', basename):
                continue
            i = int(basename[1:])
            self.latest = max(i, self.latest)
            vpath = os.path.join(self.pub_dir, basename)
            self.versions[i] = self._make_version_list(vpath)


    def _deduce_old_versions(self):
        i = 1
        self.versions = {}
        while True:
            vpath = os.path.join(self.pub_dir, 'v%d' % i)
            if not os.path.exists(vpath):
                return
            else:
                self.latest = i

            self.versions[i] = self._make_version_list(vpath)
            i += 1
            
        
    def _make_version_list(self, vpath):
        vlist = []
        for dirpath, dirnames, filenames in os.walk(vpath,
                                                    topdown=False):
            for filename in filenames:
                # Ignore files matching a regexp
                if re.match(IGNORE_FILES_REGEXP, filename):
                    continue

                filepath = os.path.join(dirpath, filename)
                drs = self.drs_tree._vtrans.filepath_to_drs(filepath)
                vlist.append((filepath, drs))
        return vlist

    def _deduce_todo(self):
        """
        Filter the drs_tree's incoming list to find new files in this
        PublisherTree.

        """

        FILTER_COMPONENTS = ['institution', 'model', 'experiment',
                             'frequency', 'realm', 'table',
                             'ensemble', 'product',
                             ]

        # Gather DRS components from the template drs instance to filter
        filter = {}
        for comp in FILTER_COMPONENTS:
            val = self.drs.get(comp, None)
            if val is not None:
                filter[comp] = val

        # Filter the incoming list
        if self.drs_tree.incoming:
            self._todo = self.drs_tree.incoming.select(**filter)
        else:
            self._todo = []

        log.info('Deduced %d incoming DRS files for PublisherTree %s' % 
                 (len(self._todo), self.drs))
                

    def _diff_file(self, filepath1, filepath2, by_tracking_id=False):
        diff_state = self.DIFF_NONE

        fp1 = os.path.realpath(filepath1)
        fp2 = os.path.realpath(filepath2)

        if fp1 != fp2:
            diff_state |= self.DIFF_PATH

        # Check files are the same size
        if _get_size(fp1) != _get_size(fp2):
            diff_state |= self.DIFF_SIZE

        # Check by tracking_id
        if by_tracking_id:
            if fp1[-3:] == fp2[-3:] == '.nc':
                if _get_tracking_id(fp1) != _get_tracking_id(fp2):
                    diff_state |= self.DIFF_TRACKING_ID

        #!TODO: what about md5sum?  This would be slow, particularly as
        #       esgpublish does it anyway.

        return diff_state



def _get_tracking_id(filename):
    import cdms2
    ds = cdms2.open(filename)
    return ds.tracking_id

def _get_size(filename):
    return os.stat(filename)[stat.ST_SIZE]




# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""


"""

#!TODO: tidy this up and review what needs documenting.

import os

import metaconfig
config = metaconfig.get_config('drslib')

##############################################################################
# Configure the location of MIP tables.  
# 
# This can be configured with the MIP_TABLE_PATH environment varliable
# or in the drslib metaconfig as 
# [drslib:tables]
# path = $MIP_TABLE_PATH
#
# You can override the location of CSV versions of the tables with the 
# "path_csv" option in metaconfig.  You can override the prefix of tables
# with the "prefix" option.
#

if 'MIP_TABLE_PATH' in os.environ:
    table_path = os.environ['MIP_TABLE_PATH']
else:
    try:
        table_path = config.get('tables', 'path')
    except:
        raise Exception("Please configure your MIP table path using MIP_TABLE_PATH or a config file")

try:
    table_path_csv = config.get('tables', 'path_csv')
except:
    table_path_csv = '%s_csv' % os.path.normpath(table_path)

# Check both paths exist
if not os.path.exists(table_path):
    raise Exception("The configured mip-table path %s does not exist" % table_path)
if not os.path.exists(table_path_csv):
    raise Exception("The mip-table CSV directory %s could not be found" % table_path_csv)


if config.has_option('tables', 'prefix'):
    table_prefix = config.get('tables', 'prefix')
else:
    table_prefix = 'CMIP5_'

##############################################################################
# Configure site-specific drs vocabulary behaviour
#

#
# Configure default values of the drs attributes
#
#!TODO: this should probably be drs_defaults.  Requires major version change. 
if config.has_section('drs'):
    drs_defaults = dict(config.items('drs'))
else:
    drs_defaults = {}

#
# Configure the model table.  You shouldn't need to change this from the
# internal default.
#
#!TODO: Check whether this is used.
try:
    model_table = config.get('tables', 'model_table')
except:
    model_table = os.path.join(os.path.dirname(__file__), 'data', 
                               'CMIP5_models.csv')

#
# Allow backward compatibility with the original version system
# Original behaviour is default
#
try:
    version_by_date = config.getboolean('versioning', 'version_by_date')
except:
    version_by_date = True

#
# Allow override of experiment names.  
# These are merged with names in the MIP tables
#
try:
    experiments = config.get('vocabularies', 'experiments').split()
except:
    experiments = []

##############################################################################
# drs_tree command defaults
#

# Default subdirectory of drs-root to scan for incoming files
DEFAULT_INCOMING = 'output'
DEFAULT_MOVE_CMD = 'mv'
try:
    move_cmd = config.get('DEFAULT', 'move-cmd')
except:
    move_cmd = DEFAULT_MOVE_CMD



#
# CMIP3 component to file/path position mapping
#

class CMIP3_DRS:
    PATH_INSTMODEL = 4
    PATH_EXPERIMENT = 0
    PATH_FREQUENCY = 2
    PATH_REALM = 1
    PATH_VARIABLE = 3
    PATH_ENSEMBLE = 5

    FILE_VARIABLE = 0
    FILE_TABLE = 1
    #FILE_SUBSET = 2
    FILE_EXTENDED = 6

#
# CMIP5 component to file/path position mapping
#

class CMIP5_DRS:
    PATH_PRODUCT = 0
    PATH_INSTITUTE = 1
    PATH_MODEL = 2
    PATH_EXPERIMENT = 3
    PATH_FREQUENCY = 4
    PATH_REALM = 5
    PATH_TABLE = 6
    PATH_ENSEMBLE = 7
    PATH_VERSION = 8
    PATH_VARIABLE = 9
    
    FILE_VARIABLE = 0
    FILE_TABLE = 1
    FILE_MODEL = 2
    FILE_EXPERIMENT = 3
    FILE_ENSEMBLE = 4
    FILE_SUBSET = 5
    FILE_EXTENDED = 6


#
# CMIP5 component to file/path position mappings for DRS <= v0.27
#

class CMIP5_DRS_OLD:
    PATH_PRODUCT = 0
    PATH_INSTITUTE = 1
    PATH_MODEL = 2
    PATH_EXPERIMENT = 3
    PATH_FREQUENCY = 4
    PATH_REALM = 5
    PATH_VERSION = 6
    PATH_VARIABLE = 7
    PATH_ENSEMBLE = 8
    
    FILE_VARIABLE = 0
    FILE_TABLE = 1
    FILE_MODEL = 2
    FILE_EXPERIMENT = 3
    FILE_ENSEMBLE = 4
    FILE_SUBSET = 5
    FILE_EXTENDED = 6

# Without version!
class CMIP5_CMOR_DRS:
    PATH_PRODUCT = 0
    PATH_INSTITUTE = 1
    PATH_MODEL = 2
    PATH_EXPERIMENT = 3
    PATH_FREQUENCY = 4
    PATH_REALM = 5
    PATH_VARIABLE = 6
    PATH_ENSEMBLE = 7
    
    FILE_VARIABLE = 0
    FILE_TABLE = 1
    FILE_MODEL = 2
    FILE_EXPERIMENT = 3
    FILE_ENSEMBLE = 4
    FILE_SUBSET = 5
    FILE_EXTENDED = 6



# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

"""
Tests compatible with nosetests

"""


import os

from drslib import cmip3, cmip5

translator = cmip3.make_translator('/')
cmip5_translator = cmip5.make_translator('cmip5')

def test_1():
    p = '/20c3m/atm/da/rsus/gfdl_cm2_0/run1/rsus_A2.19610101-19651231.nc'
    p2 = convert(p)
    assert p2 == 'cmip5/output/GFDL/CM2/20c3m/day/atmos/A2/r1/v1/rsus/rsus_A2_CM2_20c3m_r1_19610101-19651231.nc'
    
def test_listing():
    """
    Test successful conversion of part of the CMIP3 listing
    
    """
    
    fh = open(os.path.join(os.path.dirname(__file__), 'cmip3_test_ls'))
    for line in fh:
        yield convert, line.strip()
    
def convert(filepath):
    print ('%s -->' % filepath),
    drs = translator.filepath_to_drs(filepath)
    cmip5_filepath = cmip5_translator.drs_to_filepath(drs)

    print cmip5_filepath
    
    return cmip5_filepath

def test_var_underscore():
    """test_var_underscore

    Regression test found in CMIP3 when a variable contains an underscore.

    """
    p = '/1pctto2x/atm/mo/rlftoaa_co2/ipsl_cm4/run1/rlftoaa_co2_A5_1860-1869.nc'
    p2 = convert(p)

    assert p2 == 'cmip5/output/IPSL/CM4/1pctto2x/mon/atmos/A5/r1/v1/rlftoaa_co2/rlftoaa_co2_A5_CM4_1pctto2x_r1_1860-1869.nc'



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.


import logging
log = logging.getLogger(__name__)

class request_time_slice:

  def __init__(self,parent):
    self.parent=parent
    self.warning = ''

  def in_requested_time(self,start,end):
    self.is_in_requested_time = False
    ssp = self.parent.request_spec

    if ssp[0] in ['list','listrel']:
        if ssp[0] == 'listrel':
          offset = self.get_offset()
          if self.offset_status == -1:
            log.warning('failed to get offset [listrel]')
            return False
        else:
          offset = 0
        for s in ssp[1:]:
          if s[0] == 'year':
            if start-offset == s[1]:
              self.is_in_requested_time = True
              return True
          elif s[0] == 'slice':
            if start-offset <= s[2] and end-offset >= s[1]:
              self.is_in_requested_time = True
              return True

        self.requested_time_end = ssp[-1][-1]
        self.requested_time_start = ssp[1][1]
        self.is_in_requested_time = False
        return True

    elif ssp[0] == 'corres':
      ##  offset = self.get_offset()
        ##if self.offset_status == -1:
        log.warning('not ready for this yet [corres] -- need start time info %s %s ' % (tli, self.parent.request_col))
        return False

    return True

  def load_config(self):
    assert self.parent.config_exists, 'load_config: need a valid configuration file at this point'
    if not self.parent.config_loaded:
        import ConfigParser
        self.cp = ConfigParser.SafeConfigParser()
        self.cp.read( self.parent.config )
        self.parent.config_loaded = True

  def get_offset(self):
    self.load_config()
    self.offset_status = -1
    if not self.parent.model in self.cp.sections():
      self.warning += 'model [%s] not found in configuration file sections: %s, defaulting to 0 offset;;' % (self.parent.model,str(self.cp.sections()) )
      self.offset = 0
      return self.offset
    else:
      opts = self.cp.options( self.parent.model )
      if self.parent.expt == 'piControl':
        if self.parent.table == '3hr':
          if 'year_piControl_spawn_to_1pctCO2' not in opts:
            self.warning += 'start and spawn year info for 1pctCO2 not in configuration file, defaulting to 0 offset'
            self.offset= 0
            self.offset_status = 0
            return self.offset
          else:
            self.offset_status = 1
            self.offset= self.cp.get( self.parent.model, 'year_piControl_spawn_to_1pctCO2' )
            return self.offset
        else:
          if 'year_piControl_spawn_to_historical' not in opts or 'year_start_historical' not in opts:
            self.warning += 'start and spawn year info for historical not in configuration file, please amend configuration file'
            return self.offset
          else:
            self.offset_status = 1
            self.offset= self.cp.get( self.parent.model, 'year_piControl_spawn_to_historical' ) - self.cp.get( self.parent.model, 'year_stat_historical' )
            return self.offset



"""
This test module uses the functional test features of nose.

"""

import tempfile, os, shutil
import zipfile

import drslib.p_cmip5.product as p
from drslib.p_cmip5 import init
from drslib.cmip5 import make_translator
from test.gen_drs import write_listing_seq
from drslib import config

from nose import with_setup
from drs_tree_shared import test_dir

verbose = False

pc1 = None
pc2 = None
tmpdir = None
CMOR_TABLE_DIR = config.table_path
CMOR_TABLE_CSV = config.table_path_csv

def setup_module():
    global pc1, pc2, tmpdir
    tmpdir = tempfile.mkdtemp(prefix='p_cmip5-')
    shelve_dir = os.path.join(tmpdir, 'sh')

    # Example config file is in test directory
    test_dir = os.path.dirname(__file__)
    config1 = os.path.join(test_dir, 'sample_3.ini')
    config2 = os.path.join(test_dir, 'sample_4.ini')

    # Shelves are regenerated each time.  This could be optimised.
    init.init(shelve_dir,CMOR_TABLE_DIR)
    shelves = init._find_shelves(shelve_dir)

    # Create the dummy data
    z = zipfile.ZipFile(os.path.join(test_dir, 'dummy_archive.zip'))
    z.extractall(path=tmpdir)

    pc1 = p.cmip5_product(mip_table_shelve=shelves['stdo_mip'],
                          template=shelves['template'],
                          stdo=shelves['stdo'],
                          config=config1, not_ok_excpt=False)
    pc2 = p.cmip5_product(mip_table_shelve=shelves['stdo_mip'],
                          template=shelves['template'],
                          stdo=shelves['stdo'],
                          config=config2, not_ok_excpt=False)

##def teardown_module():
 ##   shutil.rmtree(tmpdir)

def do_product2(var, mip, expt,path, startyear,
            pci=None,path_output1=None,path_output2=None,verbose=False, tab='    ',selective_ads_scan=False,model='HADCM3'):

    if path != None:
      path = os.path.join(tmpdir, path)
    if path_output1 != None:
      path_output1 = os.path.join(tmpdir, path_output1)
    if path_output2 != None:
      path_output2 = os.path.join(tmpdir, path_output2)

    if pci is None:
        pci = pc1

    if startyear == -999:
       pci.find_product_ads( var, mip, expt,model,path, verbose=verbose,selective_ads_scan=selective_ads_scan)
    else:
       pci.find_product( var, mip, expt,model,path,startyear=startyear,path_output1=path_output1,path_output2=path_output2, verbose=verbose,selective_ads_scan=selective_ads_scan)

    return (pci.product, pci.rc)

def check_product3(args, kwargs, expect=None):
    r = do_product2(*args, **kwargs)
    if expect:
      if type( expect) == type( () ):
        assert r == expect, '%s :: %s' % (str(r),str(expect))
      else:
        assert r[0] == expect, '%s :: %s' % (str(r),str(expect))

def test_gen():
    for var in ['tas','pr','ua']:
        for mip in ['3hr','day']:
            if mip == '3hr':
                expect = 'output2'
            else:
                expect = 'output1'
            yield check_product3, (var, mip, 'rcp45', 'tmp/a_2005_2100', 2050), {}, expect


def test_gen2():
    yield  check_product3, ('tas', '3hr', 'rcp45', 'tmp/a_2010_2020', 2090), {}
    yield  check_product3, ('tas', '3hr', 'rcp60', 'tmp/a_2010_2020', 2090), {}
    yield  check_product3, ('hus', '6hrLev', 'historical', 'tmp/a_1930_2000', 1949), {}, ('output1', 'OK013' )
    yield  check_product3, ('hus', '6hrLev', 'historical', 'tmp/a2_1930_2000', 1949), {'selective_ads_scan':True}, ('output1', 'OK013' )

def test_gen3():
    print '3d aero field'
    yield check_product3, ( 'sconcdust', 'aero', 'rcp85', 'tmp/a_2010_2020', 2015), {}, 'output2'
    yield check_product3, ( 'rhs', 'day', 'historical', 'tmp/a_2005_2100', -999), {}, 'split'
    yield check_product3, ( 'rhs', 'day', 'historicalxxx', 'tmp/a_2005_2100', -999), {}, 'output1'
    yield check_product3, ( 'rhs', 'day', 'historical', 'tmp/a_1930_2000', -999), {}, 'split'
    yield check_product3, ( 'rhs', 'day', 'historical', 'tmp/a_1930_2000', -999), dict( pci=pc2 ), 'split'
    yield check_product3, ( 'rhs', 'day', 'piControl', 'tmp/a_1930_2000', -999), {}, 'split'
    print '------------ noVolc1980  --------------'
    yield check_product3, ( 'sconcno3', 'aero', 'noVolc1980', 'tmp/a_2005_2100', 2015 ), {'model':'HadGEM2-ES'}, ('output2', 'OK300.08')
    yield check_product3, ( 'sconcno3', 'aero', 'noVolc1980', 'tmp/a_2005_2100', 2010 ), {'model':'HadGEM2-ES'}, ('output2', 'OK300.08')
    print '------------ volcIn2010  --------------'
    yield check_product3, ( 'sconcno3', 'aero', 'volcIn2010', 'tmp/a_2005_2100', 2015 ), {'model':'HadGEM2-ES'}, ('output2', 'OK300.09')
    yield check_product3, ( 'sconcno3', 'aero', 'volcIn2010', 'tmp/a_2005_2100', 2014 ), {'model':'HadGEM2-ES'}, ('output1', 'OK300.09')

def test_gen4():
    print 'test using sample_2.ini, in which there is a 30 year offset between dating in historical and piControl'
    yield check_product3, ( 'rhs', 'day', 'piControl', 'tmp/a_1930_2000', -999 ), dict( verbose=verbose, pci=pc2 ), 'split'
    yield check_product3, ( 'tas', '3hr', 'piControl', 'tmp/a_2005_2100', -999 ),  {},'split'
    yield check_product3, ( 'tas', '3hr', 'historical', 'tmp/a_1930_2000', 1990 ),  {},'output1'
    yield check_product3, ( 'tasxx', '3hr', 'rcp45', None, 1990 ),  {},'output2'
    yield check_product3, ( 'intpdiaz', 'Omon', 'rcp45', 'tmp/a_2010_2020', 2050 ),  {},'output1'
    yield check_product3, ( 'intpp', 'Omon', 'rcp45', 'tmp/a_2010_2020', 2050 ),  {},'output1'
    yield check_product3, ( 'sconcno3', 'aero', 'piControl', 'tmp/a_2005_2100', 2040 ), {}, 'output1'
    yield check_product3, ( 'ta', '6hrPlev', 'midHolocene', 'tmp/a_1001_1050', 1040 ), {}, ('output1','OK200.01')
    yield check_product3, ( 'ta', '6hrPlev', 'midHolocene', 'tmp/a_1001_1050', 1040 ), {'path_output1':'tmp/a_1001_1030'}, ('Failed','ERR007')
    yield check_product3, ( 'ta', '6hrPlev', 'midHolocene', 'tmp/a_2005_2100', -999 ), {}, 'split'
    yield check_product3, ( 'emioa', 'aero', 'decadal2005', 'tmp/a_2005_2100', -999 ), {}, 'output1'
    yield check_product3, ( 'sconcnh4', 'aero', 'decadal2005', 'tmp/a_2005_2100', -999 ), {}, 'split'
    yield check_product3, ( 'sconcnh4', 'aero', 'decadal2005', 'tmp/a_2005_2100', 2010 ), {}, 'output2'
    yield check_product3, ( 'sconcnh4', 'aero', 'decadal2005', 'tmp/a_2005_2100', 2015 ), {}, 'output1'
    yield check_product3, ( 'sconcnh4', 'aero', 'decadal2005', 'tmp/a_2005_2100', 2045 ), {}, 'output2'
    yield check_product3, ( 'sconcnh4', 'aero', 'decadal2001', 'tmp/a_2005_2100', -999 ), {}, 'split'

def test_gen5():
    print 'cfMon, section 1'
    yield check_product3, ( 'rlu', 'cfMon', 'amip', 'tmp/a_1930_2000', 1990 ), {}, ('output1','OK300.08')
    yield check_product3, ( 'rlu', 'cfMon', 'amip', 'tmp/single', None ), {}, ('output1','OK012')

def test_gen6():
    print 'cfMon, section 2'
    yield check_product3, ( 'rlut4co2', 'cfMon', 'piControl', 'tmp/a_1930_2000', 1950 ), {}, ('output1','OK100')

def test_gen7():
    print 'cfMon, section 3'
    yield check_product3, ( 'rlu4co2', 'cfMon', 'amip', 'tmp/a_1930_2000', 1980 ), {}, ('output1','OK300.08')
    yield check_product3, ( 'rlu4co2', 'cfMon', 'piControl', 'tmp/a_1930_2000', 1950  ), {}, ('output1','OK008.2')

def test_gen8():
    print '------------ cfMon, section 4 --------------'
    yield check_product3, ( 'clisccp', 'cfMon', 'amip', 'tmp/a_1930_2000', 1980 ), {}, ('output1', 'OK300.08')
    yield check_product3, ( 'clisccp', 'cfMon', 'piControl', 'tmp/a_2005_2100', 2030 ), {}, ('output2', 'OK200.06')
    yield check_product3, ( 'clisccp', 'cfMon', 'piControl', 'tmp/a_2005_2100', 2029 ), {}, ('output1', 'OK200.06')
    yield check_product3, ( 'clisccp', 'cfMon', 'abrupt4xCO2', 'tmp/a_2010_2020', 2015 ), {}, ('output1','OK009.2')
    print '------------ 1pctCO2  --------------'
    yield check_product3, ( 'va', '6hrPlev', '1pctCO2', 'tmp/a_2010_2020', 1980 ), {'model':'HadGEM2-ES'}, ('output1', 'OK008.1')
    yield check_product3, ( 'tas', '3hr', 'abrupt4xCO2', 'tmp/a_2010_2020', 2012 ), {'model':'HadGEM2-ES'}, ('output1', 'OK009.2')
    yield check_product3, ( 'tas', '3hr', 'abrupt4xCO2', 'tmp/a_2005_2100', 2007 ), {'model':'HADCM3'}, ('output1', 'OK200.03')
    yield check_product3, ( 'tas', '3hr', 'abrupt4xCO2', 'tmp/a_2005_2100', 2016 ), {'model':'HADCM3'}, ('output2', 'OK200.03')

def test_gen9():
    print '------------ mohc, first batch --------------'
    yield check_product3, ( 'hus', '6hrLev', 'historical', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK013')
    yield check_product3, ( 'ps', '6hrLev', 'historical', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK013')
    yield check_product3, ( 'ta', '6hrLev', 'historical', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK013')
    yield check_product3, ( 'ua', '6hrLev', 'historical', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK013')
    yield check_product3, ( 'va', '6hrLev', 'historical', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK013')
    yield check_product3, ( 'psl', '6hrPlev', 'historical', 'tmp/mohc1', 1949 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK300.08')
    yield check_product3, ( 'ta', '6hrPlev', 'historical', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK300.08')
    yield check_product3, ( 'ua', '6hrPlev', 'historical', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK300.08')
    yield check_product3, ( 'va', '6hrPlev', 'historical', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK300.08')
    yield check_product3, ( 'areacella', 'fx', 'rcp85', 'tmp/mohc1', 1980 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK013')
    yield check_product3, ( 'sconcbc', 'aero', 'rcp85', 'tmp/mohc_cf/', 2005 ), {'selective_ads_scan':True, 'model':'HadGEM2-ES'}, ('output1', 'OK300.08')
    
##yield check_product3, ( 'clt', 'day', 'piControl', 'tmp/mohc1', 1979 ), {'selective_ads_scan':False, 'model':'HadGEM2-ES'}, ('output1', 'OK300.08')

##( 'rlu4co2', 'cfMon', 'piControl', startyear=2000, endyear=2000, path='./tmp/a_2010_2020', expected=('output1', 'OK008.2') )

def test_regression_1():
    """
    Bug identified during ingest.

    """    
    prefix = os.path.join(tmpdir, 'reg_1')
    filenames = [
        'clt_day_HadGEM2-ES_piControl_r1i1p1_19791201-19891130.nc',
        'clt_day_HadGEM2-ES_piControl_r1i1p1_19891201-19991130.nc',
        'clt_day_HadGEM2-ES_piControl_r1i1p1_19991201-20091130.nc',
        ]

    write_listing_seq(prefix, filenames)
    trans = make_translator(prefix)

    drs = trans.filename_to_drs(filenames[0])
    
    status = pc1.find_product(drs.variable, drs.table, drs.experiment,
                              drs.model, prefix,
                              startyear=1979)
    assert status
    assert pc1.product=='output1'

def test_regression_2():
    prefix = os.path.join(tmpdir, 'reg_2')
    filenames = [
        'areacella_fx_HadGEM2-ES_rcp85_r1i1p1.nc',
        'orog_fx_HadGEM2-ES_rcp85_r1i1p1.nc',
        'sftlf_fx_HadGEM2-ES_rcp85_r1i1p1.nc',
    ]

    write_listing_seq(prefix, filenames)
    trans = make_translator(prefix)

    for filename in filenames:
        drs = trans.filename_to_drs(filenames[0])
        status = pc1.find_product(drs.variable, drs.table, drs.experiment,
                                  drs.model, prefix)
        assert status
        assert pc1.product=='output1'


def test_drs_tree():
    """
    Test drs_tree interface to p_cmip5.
    """
    from drslib import drs_tree

    # Point drs_root at /tmp since we won't be making any upgrades.
    dt = drs_tree.DRSTree('/tmp')
    dt.set_p_cmip5(pc1)
    dt.discover(os.path.join(tmpdir, 'tmp/tas'), activity='cmip5', institute='UKMO')

    #!TODO: More robust test here.
    datasets = set(dt.pub_trees.keys())
    assert datasets == set("""
cmip5.output2.UKMO.HADCM3.piControl.3hr.atmos.3hr.r1i1p1
cmip5.output2.UKMO.HADCM3.piControl.3hr.atmos.3hr.r2i1p1
cmip5.output1.UKMO.HADCM3.piControl.3hr.atmos.3hr.r1i1p1
cmip5.output1.UKMO.HADCM3.piControl.day.atmos.day.r3i1p1
cmip5.output1.UKMO.HADCM3.piControl.3hr.atmos.3hr.r2i1p1
""".strip().split())


def test_p_cmip5_data_perms():
    """
    Regression test to detect the case when the shelve files
    are only readable by the user.

    """
    global pc1

    #!TODO: Repeating bits of setup_module() here.  Could be DRY.
    shelve_dir = os.path.join(tmpdir, 'sh')
    test_dir = os.path.dirname(__file__)
    config1 = os.path.join(test_dir, 'sample_1.ini')
    shelves = init._find_shelves(shelve_dir)
    shelve_file = shelves['stdo']


    try:
        os.chmod(shelve_file, 0400)
        # Reload shelves
        pc1 = p.cmip5_product(mip_table_shelve=shelves['stdo_mip'],
                              template=shelves['template'],
                              stdo=shelves['stdo'],
                              config=config1, not_ok_excpt=False)
        # Repeat test
        test_drs_tree()
    finally:
        os.chmod(shelve_file, 0644)
        pc1 = p.cmip5_product(mip_table_shelve=shelves['stdo_mip'],
                              template=shelves['template'],
                              stdo=shelves['stdo'],
                              config=config1, not_ok_excpt=False)
#test_p_cmip5_data_perms.__test__ = False


def check_listing(listing_file):
    prefix = os.path.join(tmpdir, 'reg_ncc')
    filenames = open(os.path.join(test_dir, listing_file)).readlines()

    write_listing_seq(prefix, filenames)
    trans = make_translator(prefix)

    for filename in filenames:
        drs = trans.filename_to_drs(filenames[0])
        status = pc1.find_product(drs.variable, drs.table, drs.experiment,
                                  drs.model, prefix)
        assert status
        assert pc1.product=='output1'

def test_regression_ncc():
    for listing in ['ncc_rcp45.ls', 'ncc_piControl.ls', 'ncc_rcp45.ls']:
        yield check_listing, listing



# BSD Licence
# Copyright (c) 2011, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.


import tempfile
import shutil
import sys, os
from glob import glob
from StringIO import StringIO
import datetime

from unittest import TestCase

import gen_drs
from drslib.drs_tree import DRSTree
from drslib.drs import path_to_drs, drs_to_path, DRS
from drslib import config

from drs_tree_shared import TestEg, TestListing

class TestEg1(TestEg):
    __test__ = True

    def setUp(self):
        super(TestEg1, self).setUp()

        gen_drs.write_eg1(self.tmpdir)

    def test_1(self):
        dt = DRSTree(self.tmpdir)
        dt.discover(self.incoming, activity='cmip5',
                    product='output1', institute='MOHC', model='HadCM3', 
                    experiment='1pctto4x', realm='atmos')

        assert len(dt.pub_trees) == 3
        k = sorted(dt.pub_trees.keys())[2]
        assert k == 'cmip5.output1.MOHC.HadCM3.1pctto4x.day.atmos.day.r3i1p1'
        pt = dt.pub_trees[k]

        assert pt.versions == {}
        assert len(pt._todo) == 15
        vars = set(x[1].variable for x in pt._todo)
        assert vars == set(('pr', 'rsus', 'tas'))
        assert pt.state == pt.STATE_INITIAL
    
    def test_2(self):
        dt = DRSTree(self.tmpdir)
        dt.discover(self.incoming, activity='cmip5',
                    product='output1', institute='MOHC', model='HadCM3')

        assert len(dt.pub_trees) == 3
        pt = dt.pub_trees.values()[0]
        assert pt.drs.realm == 'atmos'

    def test_3(self):
        dt = DRSTree(self.tmpdir)
        dt.discover(self.incoming, activity='cmip5',
                    product='output1', institute='MOHC', model='HadCM3')
        
        pt = dt.pub_trees.values()[0]
        assert pt.state == pt.STATE_INITIAL

        pt.do_version()
        assert pt.state == pt.STATE_VERSIONED
        assert len(pt.versions.keys()) == 1

        assert self.today in pt.versions.keys()

class TestEg2(TestEg):
    __test__ = True

    def setUp(self):
        super(TestEg2, self).setUp()

        gen_drs.write_eg2(self.tmpdir)

    def test_1(self):
        dt = DRSTree(self.tmpdir)
        dt.discover(self.incoming, activity='cmip5',
                    product='output1', institute='MOHC', model='HadCM3')

        assert len(dt.pub_trees) == 2
        assert set([x.drs.realm for x in dt.pub_trees.values()]) == set(['atmos', 'ocean'])

class TestEg2_1(TestEg2):
    __test__ = True

    def test_1(self):
        """Test incremental discovery"""
        dt = DRSTree(self.tmpdir)
        components = dict(activity='cmip5',
                          product='output1', institute='MOHC', model='HadCM3')
        # Call discover without incoming_dir
        dt.discover(None, **components)
        assert len(dt.pub_trees) == 0

        # Discover ocean realm
        dt.discover_incoming(self.tmpdir, realm='ocean', **components)
        assert len(dt.pub_trees) == 1

        # Discover atmos realm
        dt.discover_incoming(self.tmpdir, realm='atmos', **components)
        assert len(dt.pub_trees) == 2

        assert set([x.drs.realm for x in dt.pub_trees.values()]) == set(['atmos', 'ocean'])

    def test_2(self):
        """Test incremental discovery without calling discover() first."""
        dt = DRSTree(self.tmpdir)
        components = dict(activity='cmip5',
                          product='output1', institute='MOHC', model='HadCM3')
        assert len(dt.pub_trees) == 0

        # Discover ocean realm
        dt.discover_incoming(self.tmpdir, realm='ocean', **components)
        assert len(dt.pub_trees) == 1

        # Discover atmos realm
        dt.discover_incoming(self.tmpdir, realm='atmos', **components)
        assert len(dt.pub_trees) == 2

        assert set([x.drs.realm for x in dt.pub_trees.values()]) == set(['atmos', 'ocean'])




#!TODO: latest

#
# Test Moving from one version to another, adding a variable
#
class TestEg3(TestEg):
    __test__ = True

    def _cmor1(self):
        gen_drs.write_eg3_1(self.tmpdir)
        self.dt = DRSTree(self.tmpdir)
        self.dt.discover(self.incoming, activity='cmip5',
                         product='output1', institute='MOHC', model='HadCM3')

        (self.pt, ) = self.dt.pub_trees.values()

    def _cmor2(self):
        gen_drs.write_eg3_2(self.tmpdir)
        self.dt.discover_incoming(self.incoming, activity='cmip5',
                                  product='output1')

        
    def _exists(self, x):
        return os.path.exists(os.path.join(self.pt.pub_dir, x))
    def _listdir(self, x):
        return os.listdir(os.path.join(self.pt.pub_dir, x))
    def _listlinks(self, x):
        links = glob('%s/*' % os.path.join(self.pt.pub_dir, x))
        return [os.readlink(lnk) for lnk in links if os.path.islink(lnk)]


    def test_01(self):
        self._cmor1()
        assert len(self.pt.drs_tree.incoming) > 0

        self.pt.do_version()
        assert len(self.pt.drs_tree.incoming) == 0
        assert self.pt.count_todo() == 0
        assert len(list(self.pt.list_todo())) == 0

    def test_1(self):
        self._cmor1()
        self.pt.do_version(20100101)

        self._cmor2()
        self.pt.do_version(20100102)

        assert len(self.pt.drs_tree.incoming) == 0

        assert self._exists('files')
        assert self._exists('files/rsus_20100102')
        assert not self._exists('files/rsus_20100101')

        assert self._exists('v20100101/tas')
        assert self._exists('v20100101/pr')
        assert not self._exists('v20100101/rsus')
        assert self._exists('v20100102/rsus')

    def test_2(self):
        self._cmor1()
        self.pt.do_version(20100101)
        self._cmor2()
        self.pt.do_version(20100102)

        assert self._exists('v20100102/pr/pr_day_HadCM3_1pctto4x_r1i1p1_2000010100-2001123114.nc')

    def test_3(self):
        self._cmor1()
        assert self.pt.state == self.pt.STATE_INITIAL
        self.pt.do_version()
        assert self.pt.state == self.pt.STATE_VERSIONED
        self._cmor2()
        assert self.pt.state == self.pt.STATE_VERSIONED_TRANS
        self.pt.do_version()
        assert self.pt.state == self.pt.STATE_VERSIONED
    

    def test_4(self):
        # Check all links are to the "files" branch
        self._cmor1()
        self.pt.do_version()
        self._cmor2()
        self.pt.do_version()

        links = self._listlinks('v2/tas/r1i1p1')
        for link in links:
            assert '/files/' in link

    def test_5(self):
        self._cmor1()
        self.pt.do_version(20100101)

        latest = os.readlink(os.path.join(self.pt.pub_dir, 'latest'))
        assert latest == 'v20100101'

        self._cmor2()
        self.pt.do_version(20100102)

        latest = os.readlink(os.path.join(self.pt.pub_dir, 'latest'))
        assert latest == 'v20100102'


    def test_6(self):
        # Test differencing 2 versions

        self._cmor1()
        self.pt.do_version(20100101)
        self._cmor2()

        v1 = []
        todo = []
        for state, path1, path2 in self.pt.diff_version(20100101):
            if state == self.pt.DIFF_V1_ONLY:
                assert not 'rsus' in path1
                v1.append(path1)
            elif state == self.pt.DIFF_V2_ONLY:
                assert 'rsus' in path2
                todo.append(path2)

        assert len(v1) == 10
        assert len(todo) == 5


class TestEg3_1(TestEg3):
    """Use a separate DRSTree instance for the upgrade to test
    TestEg3 still works in this scenario.
    """

    __test__ = True

    def _cmor2(self):
        gen_drs.write_eg3_2(self.tmpdir)
        self.dt2 = DRSTree(self.tmpdir)
        self.dt2.discover_incoming(self.incoming, activity='cmip5',
                                  product='output1')
        (self.pt, ) = self.dt2.pub_trees.values()



#
# Test Moving from one version to another, updating a variable
#
class TestEg4(TestEg3):
    __test__ = True

    def _cmor1(self):
        gen_drs.write_eg4_1(self.tmpdir)
        self.dt = DRSTree(self.tmpdir)
        self.dt.discover(self.incoming, activity='cmip5',
                         product='output1', institute='MOHC', model='HadCM3')

        (self.pt, ) = self.dt.pub_trees.values()

    def _cmor2(self):
        gen_drs.write_eg4_2(self.tmpdir)
        self.dt.discover_incoming(self.incoming, activity='cmip5',
                                  product='output1')

    def test_1(self):
        self._cmor1()
        self.pt.do_version(20100101)
        self._cmor2()
        self.pt.do_version(20100102)

        assert self._exists('files')
        assert self._exists('files/tas_20100102')
        assert self._exists('v20100102/tas')


    def test_2(self):
        self._cmor1()
        self.pt.do_version(20100101)
        self._cmor2()
        self.pt.do_version(20100102)

        assert len(self._listdir('files/tas_20100101')) == 3
        assert len(self._listdir('files/tas_20100102')) == 2
        assert len(self._listdir('v20100101/tas')) == 3
        assert len(self._listdir('v20100102/tas')) == 5

    # Do test_3 from superclass
        
    # Do test_4 from superclass


    def test_6(self):
        # Test differencing 2 versions

        self._cmor1()
        self.pt.do_version(20100101)
        self._cmor2()

        v1 = []
        todo = []
        for state, path1, path2 in self.pt.diff_version(20100101):
            if state == self.pt.DIFF_V1_ONLY:
                v1.append(path1)
            elif state == self.pt.DIFF_V2_ONLY:
                todo.append(path2)

        assert len(v1) == 3
        assert len(todo) == 2


class TestEg4_1(TestEg4):
    __test__ = True

    def _cmor2(self):
        gen_drs.write_eg4_2(self.tmpdir)
        self.dt2 = DRSTree(self.tmpdir)
        self.dt2.discover_incoming(self.incoming, activity='cmip5',
                                  product='output1')

        (self.pt, ) = self.dt2.pub_trees.values()



#!FIXME: This test is meant to test removing files but needs more work.
#        cmor2 creates 2 files that are the same as cmor1.  This could
#        be interpreted as removing 3 files but do we need to check file
#        contents?
class TestEg5(TestEg4):
    __test__ = False

    def _cmor1(self):
        gen_drs.write_eg5_1(self.tmpdir)
        self.dt = DRSTree(self.tmpdir)
        self.dt.discover(self.incoming, activity='cmip5',
                         product='output1', institute='MOHC', model='HadCM3')

        (self.pt, ) = self.dt.pub_trees.values()

    def _cmor2(self):
        gen_drs.write_eg5_2(self.tmpdir)
        self.dt.discover_incoming(self.incoming, activity='cmip5',
                                  product='output1')

    # Do test1 from superclass

    def test_2(self):
        self._cmor1()
        self.pt.do_version(20100101)
        self._cmor2()
        self.pt.do_version(20100102)

        assert len(self._listdir('files/tas_20100101')) == 5
        assert len(self._listdir('files/tas_20100102')) == 2
        assert len(self._listdir('v20100101/tas')) == 5
        assert len(self._listdir('v20100102/tas')) == 5

    # Do test_3 from superclass
        
    # Do test_4 from superclass

    def test_6(self):
        # Test differencing 2 versions

        self._cmor1()
        self.pt.do_version(20100101)
        self._cmor2()

        v1 = []
        todo = []
        diff = []
        same = []
        for state, path1, path2 in self.pt.diff_version(20100101):
            if state == self.pt.DIFF_V1_ONLY:
                v1.append(path1)
            elif state == self.pt.DIFF_V2_ONLY:
                todo.append(path2)
            elif state == self.pt.DIFF_SIZE:
                diff.append(path1)
            elif state == self.pt.DIFF_NONE:
                same.append(path1)

        #!TODO: not same?  This test needs reviewing.
        assert len(v1) == 3
        assert len(same) == 2


#!FIXME: See TestEg5
class TestEg5_1(TestEg5):
    __test__ = False


    def _cmor2(self):
        gen_drs.write_eg5_2(self.tmpdir)
        self.dt2 = DRSTree(self.tmpdir)
        self.dt2.discover_incoming(self.incoming, activity='cmip5',
                                  product='output1')
        (self.pt, ) = self.dt2.pub_trees.values()






class TestSymlinks(TestEg):
    """
    Test that symlinks created during versioning are all relative.

    """
    __test__ = True

    def setUp(self):
        super(TestSymlinks, self).setUp()

        gen_drs.write_eg1(self.tmpdir)

    def test_1(self):
        dt = DRSTree(self.tmpdir)
        dt.discover(self.incoming, activity='cmip5',
                    product='output1', institute='MOHC', model='HadCM3')
        
        pt = dt.pub_trees.values()[0]
        assert pt.state == pt.STATE_INITIAL

        pt.do_version()

        for path, drs in pt.versions[pt.latest]:
            lnk = os.readlink(path)
            assert not os.path.isabs(lnk)


#----------------------------------------------------------------------------

def test_1():
    drs = path_to_drs('/cmip5', '/cmip5/output1')
    assert drs.product == 'output1'
    assert drs.institute == None

def test_2():
    drs = path_to_drs('/cmip5/', '/cmip5/output1/TEST/HadCM3/1pctto4x/day/atmos/day/r1i1p2/foo')
                      
    assert drs.institute == 'TEST'
    assert drs.experiment == '1pctto4x'
    assert drs.table == 'day'
    assert drs.ensemble == (1,1,2)

def test_3():
    drs = DRS(product='output1', institute='TEST')

    path = drs_to_path('/cmip5', drs)

    assert path == '/cmip5/output1/TEST/*/*/*/*/*/*'

def test_4():
    drs = DRS(product='output1', institute='TEST', ensemble=(1,2,3))

    path = drs_to_path('/cmip5/', drs)

    assert path == '/cmip5/output1/TEST/*/*/*/*/*/r1i2p3'

def test_5():
    drs = DRS(product='output1', institute='TEST', model='HadCM3',
              frequency='day')

    path = drs_to_path('/cmip5', drs)

    assert path == '/cmip5/output1/TEST/HadCM3/*/day/*/*/*'



# BSD Licence
# Copyright (c) 2010, Science & Technology Facilities Council (STFC)
# All rights reserved.
#
# See the LICENSE file in the source distribution of this software for
# the full license text.

import os

from drslib import cmip5, mip_table

TAMIP_TEST_PATH = os.path.join(os.path.dirname(__file__), 'TAMIP')

table_store = mip_table.MIPTableStore('%s/TAMIP_*' % TAMIP_TEST_PATH)
translator = cmip5.make_translator('tamip', table_store=table_store)

def test_gen_1():
    fh = open(os.path.join(TAMIP_TEST_PATH, 'tamip.ls'))
    for filename in fh:
        yield do, filename

def do(filename):
    drs = translator.filename_to_drs(filename)
    
    drs.version = 1
    drs.product = 'output'
    print drs
    assert drs.is_complete()
